/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.sys.testsys.modulApplUeb;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.sys.testsys.lib.MyLogger;
import de.bsvrz.sys.testsys.testsys.IMeldungFehler;
import de.bsvrz.sys.testsys.testsys.davlib.DavLib;
import de.bsvrz.sys.testsys.testsys.davlib.IDaVInfoKonsument;
import de.bsvrz.sys.testsys.testsys.davlib.RechnerDaten;

/**
 * Plugin spezifischen DaV Kommunikationsroutinen.
 * @author Chriesten
 *
 */
public class ApplUebDaV implements IDaVInfoKonsument
{
  /**
   * Aktuelle Instanz der DAV Bibliothek
   * Hilfsmethoden Ab-/Anmeldung von Attributgruppen
   */
  private DavLib m_davLib;

  /**
   * Rueckrufobjekt zum Schreiben der Betriebsmeldunge
   * wird an ApplUebPruefen weitergeleitet
   */
  private IMeldungFehler m_MeldungFehler = null;
  
  /**
   * Prüfinstanz mit Prüfthread;
   */
  private ApplUebPruefen m_applUebPruefen = null;

  /**
   * rechner auf dem StartStopp überprüft wird.
   * Sonderfälle: lokal, konfig = auf dem lokalen, auf allen im DAV konfigurierten
   */
  private String m_rechner = null;

  /**
   * Hash Map mit allen Rechnerobjekten der aktuellen DAV Verbindung
   */
  private HashMap<String, RechnerDaten> m_davRechnerMap = new HashMap<String, RechnerDaten>();

  // ---------------------------------------------
  // Konstruktor
  // ---------------------------------------------
  /**
   * Initialisierung DaV für Plugin Überprüfung System
   * @param connection Aktuelle DaV Verbindung
   * @param meldungFehler Rufe-Zurück Objekt für Betriebsmeldungen
   * @param rechner Lese Quelle für Start/Stopp rechner.prozInfo.Zustand
   */
  public ApplUebDaV( ClientDavInterface connection, IMeldungFehler meldungFehler, String rechner)
  {
    MyLogger.logln( "VERLAUF", "Konstruktur " + this.getClass().getName());
    m_MeldungFehler = meldungFehler;
    m_rechner = rechner;
    initDAV(connection);
  }

  // ---------------------------------------------

  /**
   * initialisiert die plugin DaV Aktionen.
   * @param connection DaV
   */
  public void initDAV(ClientDavInterface connection)
  {
    // Dav Lib instanziieren
    if ( connection == null )
    {
      MyLogger.logln( "ERROR", "keine Verbindung zum DAV" );
    }

    DavLib davLib = new DavLib(connection);
    m_davLib = davLib;

    // DaV Ergebnisdaten abonnieren
    davLib.abonniereDaVInfo( (IDaVInfoKonsument) this );

    // Anmeldung zum lesen der Attributgruppen
    davLib.anmeldeLeseProzessInfoList( m_davRechnerMap, m_rechner ); // Attributgruppe rechner.prozessInfo
    davLib.anmeldeLeseAngemeldeteAppl( null ); // Attributgruppe datenverteiler.angemeldeteApplikationen
    RechnerDaten.ausDavRechnerMap( m_davRechnerMap, "Liste der angemeldeten Rechner" );
  }

  /**
   * Deinitialisiert die plugin DaV Aktionen.
   */
  public void reinitDAV()
  {
    if ( m_davLib == null )
    { // Es kann nichts abgemeldet werden
      return;
    }

    // DaV Ergebnisdaten kündigen
    m_davLib.kuendigeDaVInfo( (IDaVInfoKonsument) this );

    // -----------------------------------------------
    // Abmeldung vom lesen der Attributgruppen
    // -----------------------------------------------
    abmeldeLeseApplFertigAll(); // Abmeldung Attributgruppe "angemeldeteApplikationen.applikationsFertigmeldung"
    m_davLib.abmeldeLeseAngemeldeteAppl( null ); // Attributgruppe datenverteiler.angemeldeteApplikationen
    m_davLib.abmeldeLeseProzessInfoAuto( m_davRechnerMap ); // rechner.ProzessInfo abmelden

    // PrüfungsThread beenden
    if ( m_applUebPruefen != null )
      m_applUebPruefen.set_beenden( true );
  }

  /**
   *
   * Wrapper Methode
   * siehe {@link DavLib}{@link #anmeldeLeseApplFertig(String)}
   * @param applId App-Id
   * @return boolean
   */
  public boolean anmeldeLeseApplFertig( String applId )
  {
    return m_davLib.anmeldeLeseApplFertig( applId );
  }

  /**
   *
   * Wrapper Methode
   * siehe {@link DavLib}
   * @param applId App-Id
   */
  public void abmeldeLeseApplFertig( String applId )
  {
    m_davLib.abmeldeLeseApplFertig( applId );
    return;
  }

  /**
   * Abmeldung aller angemeldeten Attributgruppe "atg.applikationsFertigmeldung"
   */
  public void abmeldeLeseApplFertigAll()
  {
    // Alle angemeldeten Attributgruppe "atg.applikationsFertigmeldung" abmelden
    ApplUebDaten applUebDaten = ApplUebDaten.getInstance();
    Map<String, ApplUebDaten.ApplDaten> appListe = applUebDaten.get_applDatenListe();
    for (Iterator<ApplUebDaten.ApplDaten> it = appListe.values().iterator(); it.hasNext();)
    {
      ApplUebDaten.ApplDaten appdat = (ApplUebDaten.ApplDaten) it.next();
      if ( (appdat.m_applId != null) && (appdat.loeZeit == null) )
        abmeldeLeseApplFertig( appdat.m_applId.toString() );
    }
    return;
  }



  /**
   * Update Methode, Auswertung (typ.)datenverteiler.(atg.)angemeldeteApplikationen
   *
   * @param dat
   *          vom DAV empfangene ResultData Instanz
   */
  public void updateAngemeldeteAppl( ResultData dat )
  {
    long vorgaengerId = dat.getObject().getId();
    if ( false )
    {
      System.out.println("ID DAV="+vorgaengerId);
    }

    Data data = dat.getData();

    // ProzessDaten.ApplListe neu belegen
    ApplUebDaten applUebDaten = ApplUebDaten.getInstance();
    Map<String, ApplUebDaten.ApplDaten> appListe = applUebDaten.get_applDatenListe();

    // Kennung in Liste fuer alle Elemente auf 0
    applUebDaten.setzeNeuzustApplDat0();

    if ( dat.hasData())
    {
      Array array = data.getArray("angemeldeteApplikation");

      int anzahl = array.getLength();

      // Aktuelle Prozesse übernehmen
      for (int j = 0; j < anzahl; j++)
      {
        Long id = array.getItem(j).getReferenceValue("applikation").getSystemObject().getId();
        String idc = id.toString();

        Boolean prozIstNeu = true;
        synchronized(appListe)
        {
          if (appListe.containsKey( idc ))
          {
            appListe.get( idc ).m_tmpNeuzust = 1; // aktiver Prozess bereits angemeldet
            prozIstNeu = false;
          }
        }

        if (prozIstNeu)
        { // aktiver Prozess neu
          ApplUebDaten.ApplDaten appdat = applUebDaten.get_applDatenInstanz();
          appListe.put( idc, appdat );
          appdat.m_applId = new Long(idc);
          appdat.m_tmpNeuzust = 2; // aktiver Prozess neu
          appdat.m_benutzer = array.getItem(j).getTextValue( "benutzer" ).getValueText();
          appdat.m_seit = array.getItem(j).getTextValue( "seit" ).getValueText();
          appdat.neuZeit = System.currentTimeMillis();
        }
      }
    }

    // Anmelden/Abmelden der Prozesse
    //-------------------------------------

    // Kopiere synchronisierte Liste (Sync Wartezeiten minieren)
    HashMap<String, ApplUebDaten.ApplDaten> appListeCpy = cpyMapSync(appListe);
    Iterator<Map.Entry<String, ApplUebDaten.ApplDaten>> it = appListeCpy.entrySet().iterator();
    while (it.hasNext())
    {
      Map.Entry<String, ApplUebDaten.ApplDaten> me = (Map.Entry<String, ApplUebDaten.ApplDaten>) it.next();
      ApplUebDaten.ApplDaten value = me.getValue();

      // Anmelden
      if (value.m_tmpNeuzust == 2)
      {
        String key = me.getKey();
        anmeldeLeseApplFertig( key.toString() );
      }

      // Abmelden
      if (value.m_tmpNeuzust == 0)
      {
        String key = me.getKey();
        abmeldeLeseApplFertig( key );
        // und aus ApplListe loeschen
        if ( (value.loeZeit == null) || (value.loeZeit == 0) )
          value.loeZeit = System.currentTimeMillis();
      }
    }
    appListeCpy = null;
  }


  /**
   * Kopiert Hashmap. Das Lesen der Quell Hashmap ist synchronisiert.
   * @param quelle
   * @return kopierte Map
   */
  private HashMap<String, ApplUebDaten.ApplDaten> cpyMapSync(Map<String, ApplUebDaten.ApplDaten> quelle)
  {
    HashMap<String, ApplUebDaten.ApplDaten> ziel = new HashMap<String, ApplUebDaten.ApplDaten>();

    Set<String> keys = quelle.keySet();
    synchronized(quelle)
    {
      for (Iterator<String> it = keys.iterator(); it.hasNext();)
      {
        String key = (String) it.next();
        ziel.put( key, quelle.get( key ) );
      }
    }

    return(ziel);
  };



  /**
   * Update Methode, Auswertung (typ.)applikation.(atg.)applikationsFertigmeldung.(asp.)standard
   *
   * @param dat
   *          vom DAV empfangene ResultData Instanz
   */
  public void updateApplFertig( ResultData dat )
  {
      // Absender feststellen (== Id der Appliaktion)
      Long absenderId = dat.getObject().getId();

      String inkarnation = null;
      String initialisiert = null;
      Boolean ist_initialisiert = null;

      // 1: Quelle ok, 2: Daten ok
      int applZustand = 0;

      if ( dat.hasData() )
      {
        applZustand = 1;
        Data data = dat.getData();
        inkarnation   = data.getTextValue( "Inkarnationsname" ).getValueText();

        initialisiert = data.getTextValue( "InitialisierungFertig" ).getValueText();
        ist_initialisiert = initialisiert.equalsIgnoreCase( "JA" ) ? true : false;
      }
      else
      {
        if ( dat.isSourceAvailable() )
          applZustand = 1;
      }
      if (false) System.out.println(applZustand);

      // Den Applikationsdatensatz aus ProzessDaten.ApplListe ergaenzen.
      ApplUebDaten applUebDaten = ApplUebDaten.getInstance();
      ApplUebDaten.ApplDaten appdat = null;
      String absenderIds = absenderId.toString();
      if ( applUebDaten.get_applDatenListe().containsKey( absenderIds ) )
      {
        appdat = applUebDaten.get_applDatenListe().get( absenderIds );
        appdat.m_prozessId = inkarnation;
        appdat.m_ist_initialisiert = ist_initialisiert;
      }
  }

  /**
   * Update Methode, Auswertung typ.rechner.prozessInfo.zustand
   *
   * @param dat
   *          vom DAV empfangene ResultData Instanz
   */
  public void updateProzessInfo( ResultData dat )
  {
    // Absender Pid feststellen
    String rechnerPid = dat.getObject().getPid();

    // StartStopp prüfen und Ergebnis in Rechnerliste aufnehmen
    // ------------------------------------------------------------------------
    // StartStopp ist Quelle für den Aspekt
    Boolean startStoppOk = dat.isSourceAvailable();

    RechnerDaten rd = m_davRechnerMap.get( rechnerPid );
    if ( rd != null )
      rd.m_pruef_erg_startStoppOk = startStoppOk;
    // ------------------------------------------------------------------------

    MyLogger.logln( "TTT_UPDINFO", "Pid=" + rechnerPid + " map="+m_davRechnerMap);

    // ProzessDaten.ApplListe neu belegen
    ApplUebDaten applUebDaten = ApplUebDaten.getInstance();
    Map<String, ApplUebDaten.StartStoppDaten> ssListe = applUebDaten.get_startStoppDatenListe();

    // Kennung in Liste fuer alle Elemente auf 0
    applUebDaten.setzeNeuzustStartStoppDat0();

    if ( dat.hasData() )
    {
      Array e = dat.getData().getArray( "Prozesse" );

      int anzahl = e.getLength();

      // Aktuelle Prozesse übernehmen
      for (int a = 0; a < anzahl; a++)
      {
        String prozessId = e.getItem( a ).getTextValue( "ProzessID" ).getValueText();

        ApplUebDaten.StartStoppDaten ssdat = null;

        Boolean prozIstNeu = true;
        synchronized(ssListe)
        {
          if (ssListe.containsKey( prozessId ))
          { // ProzessInfo Eintrag bereits bekannt
            ssdat = ssListe.get( prozessId );
            ssdat.m_tmpNeuzust = 1;
            prozIstNeu = false;
          }
        }

        if (prozIstNeu)
        { // neuer Eintrag
          ssdat = applUebDaten.get_startStoppDatenInstanz();
          ssListe.put( prozessId, ssdat );
          ssdat.m_tmpNeuzust = 2;
          ssdat.m_prozessId = prozessId;

          // Lese Aufrufparameter ; Attributliste atl.aufrufParameter.
          // Es interessiert der Programmname == Aufrufparameter ohne beginnendem -
          // ----------------------------------------------------------------------------
          ssdat.m_aufrufParameter.clear();
          int anzParams = e.getItem( a ).getArray( "AufrufParameter" ).getLength();

          for (int j = 0; j < anzParams; j++)
          {
            String aufrufParameter = e.getItem( a ).getArray( "AufrufParameter" )
                .getItem( j ).getTextValue( "AufrufparameterWert" ).getText();

            // Interessant sind nur die Argument ohne führendes '-'
            if ( !aufrufParameter.matches("^( *- *).*")  )
            {
              ssdat.m_aufrufParameter.add( aufrufParameter );
              MyLogger.logln( "AUFPARAM=", ssdat.m_rechnerPid + "(use):"+aufrufParameter );
             }
            MyLogger.logln( "AUFPARAM=", ssdat.m_rechnerPid + "(ignore):"+aufrufParameter );
          }
          // ----------------------------------------------------------------------------



        }

        ssdat.m_rechnerPid = rechnerPid;
        ssdat.m_ssName = e.getItem( a ).getTextValue( "Name" ).getValueText();
        ssdat.m_ssDatei = e.getItem( a ).getTextValue( "AusfuehrbareDatei" ).getValueText();

        // gestoppt, gestartet
        String zustand = e.getItem( a ).getTextValue( "Zustand" ).getValueText();
        if ( (ssdat.m_ssZustand == null) || (ssdat.m_ssZustand != zustand) )
        { // Zustand geändert
          ssdat.m_ssZustand_vorher = ssdat.m_ssZustand;
          ssdat.m_ssZustand = zustand;
          ssdat.m_ssZustandAendZeit = System.currentTimeMillis();
        }

        ssdat.m_ssStartZP = e.getItem( a ).getTextValue( "StartZeitpunkt" ).getValueText();
        ssdat.m_ssInitZP = e.getItem( a ).getTextValue( "InitialisierungsZeitpunkt" ).getValueText();
        ssdat.m_ssNaechstStartZP = e.getItem( a ).getTextValue( "NächsterStartZeitpunkt" ).getValueText();
      } // for (int a = 0; a < anzahl; a++)
    }

    // Wurde ein Eintrag gelöscht?
    synchronized(ssListe)
    {
      Iterator<Map.Entry<String, ApplUebDaten.StartStoppDaten>> it = ssListe.entrySet().iterator();
      while (it.hasNext())
      {
        Map.Entry<String, ApplUebDaten.StartStoppDaten> me = (Map.Entry<String, ApplUebDaten.StartStoppDaten>) it.next();
        ApplUebDaten.StartStoppDaten value = me.getValue();

        // Eintrag geloescht
        if ( (value.m_tmpNeuzust==0) && (value.m_rechnerPid.equals(rechnerPid)) )
        {
          value.m_ssZustand = "geloescht";
          if ( (value.m_loeZeit == null) || (value.m_loeZeit == 0) )
            value.m_loeZeit = System.currentTimeMillis();
        }
      }
    }
  }

  /**
   * Verarbeitet die ankommenden DaV Informationen
   * siehe Schnittstellenbeschreibung {@link IDaVInfoKonsument}{@link #neueDavInfo(ResultData)}
   */
  public void neueDavInfo( ResultData dat )
  {
    // Bitfeld. 1: ProzessInfo, 2: Angemeldete Appl, 3: Fertigmeldung
    Integer geaendert = 0;

    if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.prozessInfo" ) )
    {
      updateProzessInfo( dat );
      geaendert |= 1;
    }
    else if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.angemeldeteApplikationen" ) )
    {
      updateAngemeldeteAppl( dat );
      geaendert |= 2;
    }
    else if ( dat.getDataDescription().getAttributeGroup().getPid().equals( "atg.applikationsFertigmeldung" ) )
    {
      updateApplFertig( dat );
      geaendert |= 4;
    }

    // Muss aufgrund der angekommenen Daten eine Meldung generiert werden?
    pruefeAenderungen(geaendert);
  }

  /**
   * Prüft, ob aufgrund der ankommenden DaV-Daten Meldungen protokolliert werden müssen.
   * @param geaendert Bitfeld. 1: ProzessInfo, 2: Angemeldete Appl, 3: Fertigmeldung
   */
  private void pruefeAenderungen(Integer geaendert)
  {
    ApplUebDaten applUebDaten = ApplUebDaten.getInstance();
    Map<String, ApplUebDaten.ApplDaten> appListe = applUebDaten.get_applDatenListe();
    Map<String, ApplUebDaten.StartStoppDaten> ssListe = applUebDaten.get_startStoppDatenListe();
    MyLogger.logln("TTTT", "Aenderung ("+geaendert+")");
    MyLogger.logln("TTTT","======================================================================");
    applUebDaten.ausMap( appListe, "appListe" );
    applUebDaten.ausMap( ssListe, "ssListe" );
    MyLogger.logln("TTTT", "======================================================================");

    if ( m_applUebPruefen == null )
    {
MyLogger.logln( "YYY", "m_applUebPruefen Instanziiert");
      m_applUebPruefen = new ApplUebPruefen(m_MeldungFehler, ssListe, appListe, m_davRechnerMap);
      m_applUebPruefen.start();
    }

//    m_applUebPruefen.m_appListe = appListe;
//    m_applUebPruefen.m_ssListe = ssListe;
//    m_applUebPruefen.m_davRechnerMap = m_davRechnerMap;

    // startStoppListe überprüfen, wenn ProzessInfo oder Inkarnationsinfo geändert
    if ( ((geaendert&7)!=0) )
    {
      m_applUebPruefen.set_pruefeStartStopp( true );
    }

    // Applikationsliste prüften, wenn Applikationsinfo geändert
    if ( ((geaendert&2)==2) )
    {
      m_applUebPruefen.set_pruefeApp( true );
    }
    
    m_applUebPruefen.myNotify();
  }
}