/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.sys.testsys.modulApplUeb;


import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import de.bsvrz.sys.testsys.lib.MyLogger;
import de.bsvrz.sys.testsys.testsys.IMeldungFehler;
import de.bsvrz.sys.testsys.testsys.davlib.RechnerDaten;

/**
 * Prüft die aufgrund der DaV - Information erzeugten Informationslisten. Im Falle von Implausibilitäten werden Betriebsmeldungen protokolliert.
 *
 * @author Chriesten
 *
 */
public class ApplUebPruefen extends Thread
{
  /**
   * Rueckrufobjekt zum Schreiben der Betriebsmeldunge wurde von ModulApplUeb->ApplUebDaV weitergeleitet
   */
  private IMeldungFehler m_MeldungFehler = null;

  /**
   * aktuelle StartStoppDatenListe
   */
  public Map<String, ApplUebDaten.StartStoppDaten> m_ssListe = null;

  /**
   * aktuelle ApplDatenListe
   */
  private Map<String, ApplUebDaten.ApplDaten> m_appListe = null;

  /**
   * Hash Map mit allen Rechnerobjekten der aktuellen DAV Verbindung
   */
  public HashMap<String, RechnerDaten> m_davRechnerMap = null;

  /**
   * soll StartStopp Liste geprüft werden?
   */
  private boolean m_pruefeStartStopp = false;

  /**
   * Flag; soll sich der Thread dieser Instanz beenden?
   */
  private Boolean m_beenden = false;
  public void set_beenden( Boolean m_beenden )
  {
    this.m_beenden = m_beenden;
  }

  /**
   * Prüfung aufgrund Start/Stopp bzw. Inkarnationsänderung!
   * @param startStopp StartStopp
   */
  public void set_pruefeStartStopp( boolean startStopp )
  {
    m_pruefeStartStopp = startStopp;
  }

  /**
   * soll appListe geprüft werden?
   */
  private boolean m_pruefeApp = false;

  /**
   * Prüfung aufgrund Änderung der angemeldeten Applikationen!
   * @param app App
   */
  public void set_pruefeApp( boolean app )
  {
    m_pruefeApp = app;
  }

  /**
   * Parameter fuer thread: wartezeit in ms zwischen den Prüfungen 2 Minuten
   */
  private Long m_zyklusZeit = new Long( 30000 );

  /**
   * grace - Wartezeit. Verläuft ein Test negativ, so wird er wiederholt, falls der Testzustand und die Testzeit weniger als m_graceZeit ms auseinander liegen.
   */
  private long m_graceZeit = 60000;

  /**
   * Flag; beim ersten Zyklus soll Thread nicht warten
   */
  Boolean m_erstesMal = false;


  // -----------------------------------------------------

  // Konstruktor
  public ApplUebPruefen( IMeldungFehler meldungFehler,
      Map<String, ApplUebDaten.StartStoppDaten> ssListe,
      Map<String, ApplUebDaten.ApplDaten> appListe, HashMap<String, RechnerDaten> davRechnerMap )
  {
    m_MeldungFehler = meldungFehler;
    m_ssListe = ssListe;
    m_appListe = appListe;
    m_davRechnerMap = davRechnerMap;
  }

  /**
   * Prüft, ob aufgrund der StartStop Datenliste Meldungen protokolliert werden müssen.
   *
   */
  private void pruefeAendSSInfoListold()
  {
    // wird geprüft
    set_pruefeStartStopp( false );

    synchronized(m_ssListe)
    {
      for (Iterator<ApplUebDaten.StartStoppDaten> it = m_ssListe.values().iterator(); it.hasNext();)
      {
        ApplUebDaten.StartStoppDaten ssdat = (ApplUebDaten.StartStoppDaten) it.next();

        Boolean datensatz_loeschen = pruefeAendSSInfoDat( ssdat, m_appListe );
        if ( datensatz_loeschen )
          it.remove();
      }
    }
  }

  /**
   * Prüft, ob aufgrund der StartStop Datenliste Meldungen protokolliert werden müssen.
   *
   */
  private void pruefeAendSSInfoList()
  {
    // wird geprüft
    set_pruefeStartStopp( false );

    synchronized ( m_ssListe )
    {
      for (Iterator<ApplUebDaten.StartStoppDaten> it = m_ssListe.values().iterator(); it.hasNext();)
      {
        ApplUebDaten.StartStoppDaten ssdat = (ApplUebDaten.StartStoppDaten) it.next();

        Boolean datensatz_loeschen = pruefeAendSSInfoDat( ssdat, m_appListe );
        if ( datensatz_loeschen )
          it.remove();
      }
    }
  }


  /**
   * Prüft, ob aufgrund der Aenderungen in der Appliaktionsliste Aktionen durchgeführt werden müssen. Es werden die als gelöschten markierten Einträge aus der
   * Liste entfernt.
   *
   */
  private void pruefeAendAppList()
  {
    // Wird geprüft
    set_pruefeApp( false );

    // Suche Zeile mit gelöschten Daten
    synchronized ( m_appListe )
    {
      for (Iterator<ApplUebDaten.ApplDaten> it = m_appListe.values().iterator(); it.hasNext();)
      {
        ApplUebDaten.ApplDaten appdat = (ApplUebDaten.ApplDaten) it.next();
        if ( (appdat.loeZeit != null) && (appdat.loeZeit > 0) )
          it.remove();
      }
    }
  }

  /**
   * Synchronisiert {@link #pruefeAendSSInfoDat(de.bsvrz.sys.testsys.modulApplUeb.ApplUebDaten.StartStoppDaten, Map)} auf ssdat. Parameter und return siehe
   * obigen Link.
   */
  private Boolean pruefeAendSSInfoDat_sync( ApplUebDaten.StartStoppDaten ssdat,
      Map<String, ApplUebDaten.ApplDaten> appListe )
  {
    synchronized ( ssdat )
    {
      return (pruefeAendSSInfoDat( ssdat, appListe ));
    }
  }

  /**
   * Liefert true, falls der übergebene Block nicht geprüft werden soll.
   * Private Hilfsklasse zu {@link ApplUebPruefen}{@link #pruefeAendSSInfoDat(de.bsvrz.sys.testsys.modulApplUeb.ApplUebDaten.StartStoppDaten, Map)}
   * DaV Transmitter + Konfiguration tauchen in AppListe nicht auf.
   * Sie dürfen nicht geprüft werden.
   * @param ssdat Liste mit einem StartStopp Datensatz
   * @return true: nicht prüfen, false sonst
   */
  private Boolean nichtPruefen( ApplUebDaten.StartStoppDaten ssdat )
  {
    Boolean nichtPruefen = false;
    String meineExe = ssdat.m_ssDatei.toUpperCase();
    String vglPat = "(.*[.]TRANSMITTER.*)|(.*[.]CONFIGURATIONAPP.*)";
    if ( meineExe.matches(vglPat) )
    {
      nichtPruefen = true;
      return(nichtPruefen);
    }

    for (Iterator<String> it = ssdat.m_aufrufParameter.iterator(); it.hasNext();)
    {
      meineExe = (String) it.next();

      if ( meineExe.matches(vglPat) )
      {
        nichtPruefen = true;
        return(nichtPruefen);
      }
    }
    return(nichtPruefen);
  }

  /**
   * Prüft, ob aufgrund des geänderten StartStop Datensatzes Meldungen protokolliert werden müssen.
   *
   * @param ssdat
   *          Liste mit einem StartStopp Datensatz
   * @param appListe
   *          Liste mit den Appliaktionsdaten
   * @return true: diesen Datensatz aus StartStopp Liste löschen
   */
  private Boolean pruefeAendSSInfoDat( ApplUebDaten.StartStoppDaten ssdat,
      Map<String, ApplUebDaten.ApplDaten> appListe )
  {
    MyLogger.logln( "TTT_PP", "pruefeAendSSInfoDat gerufen " + ssdat.m_prozessId );

    // Datum überhaupt prüfen?
    if ( (ssdat.m_nichtPruefen == null) )
    {
      ssdat.m_nichtPruefen = nichtPruefen( ssdat );
    }
    if ( ssdat.m_nichtPruefen )
    {
      MyLogger.logln( "TTT_PP", "nicht pruefen " + ssdat.m_prozessId );
      return( (ssdat.m_loeZeit != null) && (ssdat.m_loeZeit>0) );
    }

    ApplUebDaten.ApplDaten appdat = get_appdatInkarnation( ssdat.m_prozessId, appListe );

    Boolean AnwendungOk = (appdat != null) && (appdat.loeZeit == null);

    // merken. Möglichkeit der Restauration.
    Long ssZustandAendZeit_gemerkt = ssdat.m_ssZustandAendZeit;
    ssdat.m_ssZustandAendZeit = null;

    // auszugebende Meldung
    String mess = null;

    // diesen Datensatz aus Prozessinfoliste löschen?
    Boolean datensatz_loeschen = false;

    int pruef_erg = 0; // 0: keine Prüfung, 1: ok, 3: Fehler

    MyLogger.logln( "PRSTAT", "status="
        + ssdat.m_ssZustand
        + " ,pid="
        + ssdat.m_prozessId
        + " ,quelle=pruefeSSInfo" );

    E_ZUSTAND e_zustand;
    e_zustand = E_ZUSTAND.getEnum( ssdat.m_ssZustand );

    if ( e_zustand == null )
    { // unbekannter Zustand!!
      MyLogger.logln( "ERROR", "Zustand "
          + ssdat.m_ssZustand
          + " unbekannt. "
          + this.getClass().getName() );
      return (datensatz_loeschen);
    }

    switch ( e_zustand )
    {
      case ANGELEGT:
      break;

      case STARTENWARTEN:
      break;

      case STARTEN:
      break;

      case GESTARTET:
        if ( AnwendungOk )
        {
          mess = "Anwendung <" + ssdat.m_prozessId + "> gestartet";
          pruef_erg = 1;
        }
        else
        {
          // wenn die Startmeldung zu jung, Prüfung später wiederholen
          if ( pruefeGraceZeit( m_graceZeit, ssZustandAendZeit_gemerkt, ssdat ) )
            return (datensatz_loeschen);

          mess = "Anwendung <" + ssdat.m_prozessId + "> gibt keine Rückmeldung";
          pruef_erg = 3;
        }
      break;

      case INITIALISIERT:
        if ( AnwendungOk && appdat.m_ist_initialisiert )
        {
          mess = "Anwendung <" + ssdat.m_prozessId + "> initialisiert";
          pruef_erg = 1;
        }
        else
        {
          // wenn die Initalisierungsmeldung zu jung, Prüfung später wiederholen
          if ( pruefeGraceZeit( m_graceZeit, ssZustandAendZeit_gemerkt, ssdat ) )
            return (datensatz_loeschen);

          pruef_erg = 3;
          if ( !AnwendungOk )
            mess = "Anwendung <" + ssdat.m_prozessId + "> gibt keine Rückmeldung";
          else
            mess = "Anwendung <" + ssdat.m_prozessId + "> gestartet aber nicht initialisiert";
        }
      break;

      case GESTOPPT:
        if ( !AnwendungOk )
        {
          pruef_erg = 1;
          mess = "Anwendung <" + ssdat.m_prozessId + "> gestoppt";
        }
        else
        {
          // wenn die Stoppmeldung zu jung, Prüfung später wiederholen
          if ( pruefeGraceZeit( m_graceZeit, ssZustandAendZeit_gemerkt, ssdat ) )
            return (datensatz_loeschen);

          pruef_erg = 3;
          mess = "(gestoppte) Anwendung <" + ssdat.m_prozessId + "> akitv";
        }
      break;

      case STOPPEN:
      break;

      case INTERVALL:
        if ( !AnwendungOk )
        {
          pruef_erg = 1;
          mess = "Die Anwendung <"
              + ssdat.m_prozessId
              + "> wird um "
              + ssdat.m_ssNaechstStartZP
              + " gestartet";
        }
        else
        {
          pruef_erg = 0;
        }
      break;

      case FEHLER:
        pruef_erg = 3;
        mess = "Fehler in Anwendung <" + ssdat.m_prozessId + ">";
      break;

      case STOPPENWARTEN:
      break;

      case GELOESCHT:
        // StartStopp Eintrag (rechner.ProzessInfo) wurde gelöscht
        if ( AnwendungOk )
        {
          pruef_erg = 3;
          mess = "StartStopp Eintrag für <"
              + ssdat.m_prozessId
              + "> gelöscht. Anwendung noch aktiv.";
        }
        else
        {
          pruef_erg = 1;

          // diesen Eintrag löschen!
          datensatz_loeschen = true;

          // Gab es schon Fehler aufgrund der Löschung?
          if ( (ssdat.m_pruef_ssZustand == ssdat.m_ssZustand) && (ssdat.m_pruef_erg == 3) )
            mess = "Anwendung <" + ssdat.m_prozessId + "> beendet.";
          else
          // uninteressant???
          {
            mess = "Start/Stopp Eintrag für <"
                + ssdat.m_prozessId
                + "> gelöscht. Anwendung beendet.";
          }
        }
      break;

      default:
      break;
    }
    ;
    MyLogger.logln( "TTT_PPP", "pruefeAendSSInfoDat ausgewertet: "
        + ssdat.m_prozessId
        + ", pruefErgLokal="+pruef_erg
        + ", zustand="+ssdat.m_ssZustand
        + " ,prZustand="
        + ssdat.m_pruef_ssZustand
        + " ,prerg="
        + ssdat.m_pruef_erg );

    if ( (pruef_erg == 0) )
    {
      return (datensatz_loeschen); // Keine Prüfung, keine Meldung generieren.
    }

    if ( (mess == null) )
    { // // keine Meldung generieren
      return (datensatz_loeschen);
    }

    // Keine Änderung gegenüber der letzten Prüfung. Keine Meldung.
    if ( (ssdat.m_pruef_erg == pruef_erg)
        && (ssdat.m_pruef_ssZustand.equalsIgnoreCase( ssdat.m_ssZustand )) )
    {
      return (datensatz_loeschen);
    }

    ssdat.m_pruef_ssZustand = ssdat.m_ssZustand;
    int pruef_erg_alt = ssdat.m_pruef_erg;
    ssdat.m_pruef_erg = pruef_erg;

    // Gutmeldungen nur protokollieren, falls letztes mal Fehler
    // Zum besseren Verlauftest auskommentiert ( && false )
    if ( pruef_erg == 1 )
    { // Ok. nur Meldung, wenn letztes mal Fehler
      if ( (pruef_erg_alt != 3) && (false) )
        return (datensatz_loeschen);
    }

    // Fehlermeldungen nur protokollieren, wenn StartStopp ok
    // Zum besseren Verlauftest auskommentiert ( && false )
    if ( (pruef_erg == 3) && (!startStoppOk( ssdat.m_rechnerPid )) && (false) )
      return (datensatz_loeschen);

    m_MeldungFehler.fehlerGefunden( mess, pruef_erg, this.getClass().getName() );
    return (datensatz_loeschen);
  }

  /**
   * prüft "grace" - Zeit im Fehlerfall ; Hilfsfunktion zu {@link ApplUebPruefen}{@link #pruefeAendSSInfoDat(de.bsvrz.sys.testsys.modulApplUeb.ApplUebDaten.StartStoppDaten, Map)}.
   * Im Falle von Soll/Ist - Fehlern wird diese Routine gerufen. Liegt die Sollbedingung weniger als graceZeit ms an, so wird die Prüfung annulliert und die
   * Routine liefert true.
   *
   * @param graceZeit
   *          "grace" Zeit, siehe Funktionbeschreibung.
   * @param sollZustZeit
   *          Empfangszeit Soll Zustand (in ms)
   * @param ssdat
   *          Prüfdatensatz
   * @return true: Prüfungsroutine abbrechen. "grace"-Zeit noch nicht abgelaufen, false sonst
   */
  private boolean pruefeGraceZeit( long graceZeit, Long sollZustZeit,
      ApplUebDaten.StartStoppDaten ssdat )
  {
    Boolean bret;

    // frühester Prüfzeitpunkt (in ms seit 19:00)
    Long pruefWarteZeit = sollZustZeit == null ? null : sollZustZeit
        + graceZeit
        - System.currentTimeMillis();

    if ( (pruefWarteZeit != null) && (pruefWarteZeit > 0) )
    { // Prüfung wiederholen

      ssdat.m_ssZustandAendZeit = sollZustZeit; // Datensatz restaurieren
      bret = true; // Preufung des Satzes abbrechen

      // Start Stopp Liste im nächsten Zyklus nocheinmal prüfen
      m_pruefeStartStopp = true;
      MyLogger.logln( "TTT_PPGRACE:", "grace=" + bret + "pruefWarteZeit=" + pruefWarteZeit / 1000 );
    }
    else
      bret = false;

    return (bret);
  }

  /**
   * Liefert den Appliaktionsdatensatz mit der uebergebenen Inkarnation (ProzessId), null wenn nicht vorhanden
   *
   * @param inkarnation
   *          zu suchende Inkarnation
   * @param appListe
   *          Liste mit den Appliaktionsdaten
   * @return gefundener Satz oder null
   */
  private ApplUebDaten.ApplDaten get_appdatInkarnation( String inkarnation,
      Map<String, ApplUebDaten.ApplDaten> appListe )
  {
    if ( inkarnation == null || appListe == null )
      return null;

    synchronized(appListe)
    {
      for (Iterator<ApplUebDaten.ApplDaten> it = appListe.values().iterator(); it.hasNext();)
      {
        ApplUebDaten.ApplDaten appdat = (ApplUebDaten.ApplDaten) it.next();
        if ( inkarnation.equalsIgnoreCase( appdat.m_prozessId ) )
        {
          return appdat;
        }
      }
    }
    return null;
  }

  /**
   * Prüft, ob StartStopp auf dem übergebene Rechner läuft.
   * Enthält die Rechnerliste genau einen Eintrag, dann wird Start/Stopp Zustand dieses Eintrags
   * zurückgeliefert.
   *
   * @param rechnerPid typ.rechner ObjektId
   * @return true, false, null: StartStopp ok, nicht ok, kann nicht entschieden werden
   */
  private Boolean startStoppOk( String rechnerPid )
  {
    Boolean startStoppOk = null;
    if ( rechnerPid == null || m_davRechnerMap == null || (m_davRechnerMap.size() == 0) )
      return null;

    RechnerDaten rd = m_davRechnerMap.get( rechnerPid );

    if ( (rd!=null ) )
      startStoppOk = rd.m_pruef_erg_startStoppOk;

    return startStoppOk;
  }

  /**
   * Aufzaehlung ; Kopie der gleichnamigen StartStopp Aufzählung und Erweiterung um den Zustand GELOESCHT SkriptVewEnums.java
   *
   * @author Chriesten
   *
   */
  private enum E_ZUSTAND
    {
      /**
       * Inkarnation wurde angelegt
       */
      ANGELEGT("angelegt"),

      /**
       * Inkarnation wartet auf Erfüllung der Startbedingung
       */
      STARTENWARTEN("warte Startbedingung"),

      /**
       * Inkarnation wird gestartet
       */
      STARTEN("starten"),

      /**
       * Inkarnation ist gestartet
       */
      GESTARTET("gestartet"),

      /**
       * Inkarnation ist initialisiert (Rückmeldung von Applikation)
       */
      INITIALISIERT("initialisiert"),

      // TODO Schreibfehler Stoppbedinung entfernen wenn neuer DaV wieder läuft
      /**
       * Inkarnation wartet auf Erfüllung der Stoppbedingung
       */
      STOPPENWARTEN("warten Stoppbedinung"),

      /**
       * Inkarnation wird gestoppt
       */
      STOPPEN("stoppen"),

      /**
       * Inkarnation ist gestoppt
       */
      GESTOPPT("gestoppt"),

      /**
       * Inkarnation wartet auf den nächsten zyklischen Neustart
       */
      INTERVALL("warte Intervall"),

      /**
       * beim Start der Inkarnation trat ein Fehler auf
       */
      FEHLER("Fehler"),

      /**
       * Prozessinfo Eintrag wurde gelöscht Erweiterung gegenüber E_ZUSTAND in Start/Stopp!
       */
      GELOESCHT("geloescht");

      private String m_text;

      E_ZUSTAND( String text )
      {
        m_text = text;
      }

      public String getText()
      {
        return m_text;
      }

      public static E_ZUSTAND getEnum( String text )
      {
        for (E_ZUSTAND z : E_ZUSTAND.values())
          if ( z.getText().equalsIgnoreCase( text ) )
            return (z);

        return null;
      }
    };

  /**
   * Wrapper fuer wait
   *
   * @param warteZeit
   */
  private synchronized void mywarte( Long warteZeit )
  {
    if ( warteZeit == null )
      return;

    if ( m_erstesMal )
    {
      m_erstesMal = false;
      return;
    }

    try
    {
      wait( m_zyklusZeit );
    }
    catch ( InterruptedException e )
    {
      interrupt();
    }
    catch ( Exception e )
    {
      MyLogger.logln( "ERROR", "run() in " + this.getClass().toString() + ":" + e.getMessage() );
      e.printStackTrace();
    }
  }

  /**
   * Methode zum Unterbrechen der wait-Zyklen im Prüfe Thread
   */
  public synchronized void myNotify()
  {
    this.notify ();
    MyLogger.logln( "TTT_APPLUEB", "notify gerufen");
  }


  /**
   * Prüft Änderungen in pruefeAendSSInfoList()
   */
  @Override
  public void run()
  {
    MyLogger.logln( "TTT_APPLUEBPRUEF", "Thread in ApplUebPruefen gestartet" );

    while ( !m_beenden )
    {
      int pruefmod = 0;

      mywarte( m_zyklusZeit );

      if ( m_pruefeStartStopp )
      {
        pruefeAendSSInfoList();
        pruefmod |= 1;
      }

      if ( m_pruefeApp )
      {
        pruefeAendAppList();
        pruefmod |= 2;
      }
      MyLogger.logln( "TTT_APPLUEBPRUEF", "Threadschleife in ApplUebPruefen beendet:"
          + pruefmod
          + " ,"
          + m_zyklusZeit );
    }

    MyLogger.logln( "TTT_APPLUEBPRUEF", "Thread in ApplUebPruefen beendet" );
  }

  /**
   * Testroutine fuer das lokale Modul. Wird fuer verschiedene Tests im Laufe des Entwicklungs- und Wartungsprozesses verwendet.
   * 
   * @param args
   *          Testargumente
   */
  public static void main( String[] args )
  {
    Long aendzeit = System.currentTimeMillis() - 2000;
    Long elapsed = aendzeit == null ? null : (System.currentTimeMillis() - aendzeit) / 1000;
    System.out.println( elapsed );

  }

}
